# -*- coding: utf-8 -*-
from abc import ABCMeta
from itertools import combinations, combinations_with_replacement, product
from typing import Any, Iterable, List, Tuple, Union

import networkx as nx
import numpy as np

from .utils.functions import (
    chunks,
    compute_wl_hashes,
    distribute_function,
    flatten_lists,
    generate_random_strings,
)
from .utils.validation import check_wl_input

1


class WeisfeilerLehmanKernel:
    """Weisfeiler-Lehmann kernel"""

    def __init__(
        self,
        n_jobs: int,
        n_iter: int = 3,
        node_label: str = "residue",
        normalize: bool = False,
        biased: bool = False,
        verbose: bool = False,
    ):
        self.n_iter = n_iter
        self.node_label = node_label
        self.normalize = normalize
        if n_jobs is not None:
            self.n_jobs = int(n_jobs)
        else:
            self.n_jobs = None
        self.biased = biased
        self.verbose = verbose

    def compute_gram_matrix(
        self, X: List[nx.Graph], Y: Union[List[nx.Graph], None] = None
    ) -> np.ndarray:
        def parallel_dot_product(
            lst: Iterable,
        ) -> Iterable:  # pragma: no cover
            res = list()
            for x in lst:
                res.append(
                    {
                        list(x.keys())[0]: [
                            list(x.values())[0][0],
                            dot_product(list(x.values())[0][1]),
                        ]
                    }
                )
            return res

        def dot_product(dicts: Tuple) -> int:  # pragma: no cover

            running_sum = 0
            # 0 * x = 0 so we only need to iterate over common keys
            for key in set(dicts[0].keys()).intersection(dicts[1].keys()):
                running_sum += dicts[0][key] * dicts[1][key]
            return running_sum

        def handle_hashes_single_threaded(
            X: Iterable[nx.Graph],
        ) -> Iterable[nx.Graph]:  # pragma: no cover
            X_hashed = list()
            for g in X:
                X_hashed.append(
                    compute_wl_hashes(
                        g, node_label=self.node_label, n_iter=self.n_iter
                    )
                )
            return X_hashed

        check_wl_input(X)
        if Y is not None:
            check_wl_input(Y)

        if Y == None:  # pragma: no cover
            Y = X

        if Y == X and self.n_jobs is not None:
            X_hashed = distribute_function(
                compute_wl_hashes,
                X,
                self.n_jobs,
                show_tqdm=self.verbose,
                tqdm_label="Compute hashes of X",
                node_label=self.node_label,
                n_iter=self.n_iter,
            )
            Y_hashed = X_hashed
        elif X == Y and self.n_jobs is None:
            X_hashed = handle_hashes_single_threaded(X)
            Y_hashed = X_hashed
        elif X != Y and self.n_jobs is None:
            X_hashed = handle_hashes_single_threaded(X)
            Y_hashed = handle_hashes_single_threaded(Y)
        elif X != Y and self.n_jobs is not None:
            X_hashed = distribute_function(
                compute_wl_hashes,
                X,
                n_jobs=self.n_jobs,
                show_tqdm=self.verbose,
                node_label=self.node_label,
                tqdm_label="Compute hashes of X",
                n_iter=self.n_iter,
            )
            Y_hashed = distribute_function(
                compute_wl_hashes,
                Y,
                n_jobs=self.n_jobs,
                show_tqdm=self.verbose,
                tqdm_label="Compute hashes of Y",
                node_label=self.node_label,
                n_iter=self.n_iter,
            )

        # It's faster to process n_jobs lists than to have one list and
        # dispatch one item at a time.
        if not self.biased and X == Y:
            iters_data = list(list(combinations(X_hashed, 2)))
            iters_idx = list(combinations(range(len(X_hashed)), 2))
        elif self.biased and X == Y:
            iters_data = list(list(combinations_with_replacement(X_hashed, 2)))
            iters_idx = list(
                combinations_with_replacement(range(len(X_hashed)), 2)
            )
        else:
            iters_data = list(list(product(X_hashed, Y_hashed)))
            iters_idx = list(
                product(range(len(X_hashed)), range(len(Y_hashed)))
            )

        keys = generate_random_strings(10, len(flatten_lists(iters_data)))
        iters = [
            {key: [idx, data]}
            for key, idx, data in zip(keys, iters_idx, iters_data)
        ]
        if self.n_jobs is not None:
            iters = list(chunks(iters, self.n_jobs,))
            matrix_elems = flatten_lists(
                distribute_function(
                    parallel_dot_product,
                    iters,
                    self.n_jobs,
                    show_tqdm=self.verbose,
                    tqdm_label="Compute dot products",
                )
            )
        else:
            matrix_elems = parallel_dot_product(iters)

        K = np.zeros((len(X_hashed), len(Y_hashed)), dtype=int)
        for elem in matrix_elems:
            coords = list(elem.values())[0][0]
            val = list(elem.values())[0][1]
            K[coords[0], coords[1]] = val
        if X == Y:
            # mirror the matrix along diagonal
            K = np.triu(K) + np.triu(K, 1).T
        return K



# -*- coding: utf-8 -*-
import contextlib
from collections import Counter
from itertools import product
from string import ascii_letters
from typing import Any, Callable, Dict, Iterable, List

import joblib
import networkx as nx
from joblib import Parallel, delayed
from tqdm import tqdm

from .exception import UniquenessError


def generate_random_strings(string_length: int, n_strings: int) -> List[str]:
    unique_strings = list()
    for idx, item in enumerate(product(ascii_letters, repeat=string_length)):
        if idx == n_strings:
            break
        unique_strings.append("".join(item))

    if len(unique_strings) != n_strings:
        raise UniquenessError(
            f"Cannot generate enough unique strings from given string length "
            "of {string_length}. Please increase the string_length to continue."
        )
    return unique_strings


def distribute_function(
    func: Callable,
    X: Iterable,
    n_jobs: int,
    tqdm_label: str = None,
    total: int = 1,
    show_tqdm: bool = True,
    **kwargs,
) -> Any:
    """Simply distributes the execution of func across multiple cores to process X faster"""
    if total == 1:
        total = len(X)
    if show_tqdm:
        with tqdm_joblib(tqdm(desc=tqdm_label, total=total)) as progressbar:
            Xt = Parallel(n_jobs=n_jobs)(delayed(func)(x, **kwargs) for x in X)
    else:
        Xt = Parallel(n_jobs=n_jobs)(delayed(func)(x, **kwargs) for x in X)
    return Xt


def flatten_lists(lists: list) -> list:
    """Removes nested lists"""
    result = list()
    for _list in lists:
        _list = list(_list)
        if _list != []:
            result += _list
        else:  # pragma: no cover
            continue
    return result


def chunks(lst, n):
    """returns lst divided into n chunks approx. the same size"""
    k, m = divmod(len(lst), n)
    return (
        lst[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n)
    )


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar.

    Code stolen from https://stackoverflow.com/a/58936697
    """

    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()


def compute_wl_hashes(G: nx.Graph, node_label: str, n_iter: int) -> Dict:
    hash_iter_0 = dict(Counter(list(dict(G.nodes(node_label)).values())))
    hashes = dict(
        Counter(
            flatten_lists(
                list(
                    nx.weisfeiler_lehman_subgraph_hashes(
                        G, node_attr=node_label, iterations=n_iter,
                    ).values()
                )
            )
        )
    )
    return hashes | hash_iter_0



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastwlk', 'fastwlk.utils']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.6.3,<3.0.0',
 'numpy>=1.22.1,<2.0.0',
 'pandas>=1.4.0,<2.0.0',
 'tqdm>=4.62.3,<5.0.0']

setup_kwargs = {
    'name': 'fastwlk',
    'version': '0.1.0.dev10',
    'description': 'fastwlk is a Python package that implements a fast version of the Weisfeiler-Lehman kernel.',
    'long_description': '=============================\nFastWLK\n=============================\n\nfastwlk is a Python package that implements a fast version of the Weisfeiler-Lehman kernel.\n',
    'author': 'Philip Hartout',
    'author_email': 'philip.hartout@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
